/*
This is code taken from "Mastering Algorithms with C" with a couple of simple macros of my own.
*/

/*****************************************************************************
*                                                                            *
*  ------------------------------- dlist.h --------------------------------  *
*                                                                            *
*****************************************************************************/

#ifndef DLIST_H
#define DLIST_H

#include <stdlib.h>

/*****************************************************************************
*                                                                            *
*  Define a structure for doubly-linked list elements.                       *
*                                                                            *
*****************************************************************************/

typedef struct DListElmt_
{

  void *data;
  struct DListElmt_ *prev;
  struct DListElmt_ *next;

}
DListElmt;

/*****************************************************************************
*                                                                            *
*  Define a structure for doubly-linked lists.                               *
*                                                                            *
*****************************************************************************/

typedef struct DList_
{

  int size;

  int (*match) (const void *key1, const void *key2);
  void (*destroy) (void *data);

  DListElmt *head;
  DListElmt *tail;

}
DList;

/*****************************************************************************
*                                                                            *
*  --------------------------- Public Interface ---------------------------  *
*                                                                            *
*****************************************************************************/

void dlist_init (DList * list, void (*destroy) (void *data));

void dlist_destroy (DList * list);

void dlist_ins_next (DList * list, DListElmt * element, const void *data);

void dlist_ins_prev (DList * list, DListElmt * element, const void *data);

bool dlist_remove(DList * list, DListElmt * element, void **data);

void dlist_append (DList * list, const void *data);

#define dlist_size(list) ((list)->size)

#define dlist_head(list) ((list)->head)

#define dlist_tail(list) ((list)->tail)

#define dlist_data(element) ((element)->data)

#define dlist_next(element) ((element)->next)

#define dlist_iter(listname,element) for(element = dlist_head(listname);element !=NULL;element=element->next)

#endif
