#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "dlist.h"


typedef struct _probe_entry
{
  float r;
  float g;
  float b;
  float u;
  float v;
}
ProbeEntry;


typedef struct _probe
{
  ProbeEntry *map;
  int usize;
  int vsize;
  int nsamples;
}
ProbeData;


typedef struct _vector
{
  float x;
  float y;
  float z;
}
Vector;


typedef struct _intimage
{
  int *data;
  int usize;
  int vsize;
  int size;
}
IntImage;

typedef struct _doubleimage
{
  double *data;
  int usize;
  int vsize;
  int size;
}
DoubleImage;


typedef struct _darray
{
  double *data;
  int size;
}
DoubleArray;

typedef struct _light
{
  double pos[3];
  double color[3];
  double area;
  int index;
  //debugging info
  int n;
  int old;
}
Light;

const double EPS = 1.0e-12;

// Function prototypes

void alloc_probe (ProbeData * probe);
void free_probe (ProbeData * probe);

void alloc_intimage (IntImage * image);
void free_intimage (IntImage * image);

IntImage *new_image (IntImage * image);
void alloc_doubleimage (DoubleImage * image);
void free_doubleimage (DoubleImage * image);

bool read_probe (char *filename, ProbeData * probe);

float find_mask (ProbeData * probe, float lambda, IntImage * mask);
void get_pos (ProbeEntry * node, double *pos);
float get_area (ProbeEntry * entry);
double patch_area (ProbeData * probe, int i);

void init_intimage (IntImage * image);

// morph stuff

void erode (IntImage * image, IntImage * output);
void dilate (IntImage * image, IntImage * output);
void clean_mask (IntImage * image, IntImage * output);

int bwlabel (IntImage * image, IntImage * mask);

void shmoys (ProbeData * patch, int number_of_lights, DList * lightlist);

void print_rle (DList * list);

void lights_ascii_writer (DList * lightlist, int nregions, int count,
                         char *filename);
void lights_binary_writer (DList * lightlist, int nregions, int count,
                          char *filename);

bool subset_intimage (IntImage * a, IntImage * b, int i, int j);

DoubleArray *calculate_svalue (ProbeData * probe, double lambda, double alpha,
                               double beta, double dw);

double patchs (ProbeData * probe, IntImage * mask, double alpha, double beta,
               double dw, int label);

void copy_intimage (IntImage * a, IntImage * b);
