#include "probe.h"
#include "dlist.h"

/*

This function implements connected components analysis. 
It proceeds by doing RLE compression, or perhaps I should say 
connected components in one dimension along each column ( well since we are using the matlab convention of storing stuff columnwise, and then joining them into sets or something like that

It returns the number of components as its return value.

Another thing is that I should convert that iArray into a two dimensional thing
where it is actually a two dimensional array thingie
*/


typedef struct rle_node_
{
  int col;                      // column number
  int min;                      // start row
  int max;                      // end ro
}
rlenode;

inline int
index (int i, int j, int u)
{
  return u * j + i;
}



inline int
intersect (int a, int b, int c, int d)
{
  int p, q;
  p = c;
  if (a < c)
    p = a;
  q = d;
  if (b > d)
    q = b;

  return ((q - p) <= (b - a + d - c + 1));
}

inline int
cmp_rle (rlenode * node1, rlenode * node2)
{
  return ((node1->col < node2->col)
          || ((node1->col == node2->col) && (node1->min > node2->min)));
}


int
merge_rle (DList * list1, DList * list2, DList * output)
{
  DListElmt *node1;
  DListElmt *node2;

  int count1;
  int count2;

  count1 = dlist_size (list1);
  count2 = dlist_size (list2);

  node1 = dlist_head (list1);
  node2 = dlist_head (list2);

  while ((node1 != NULL) && (node2 != NULL))
  {
    if (cmp_rle
        ((rlenode *) dlist_data (node1), (rlenode *) dlist_data (node2)))
    {
      dlist_ins_next (output, dlist_tail (output), dlist_data (node1));
      count1--;
      node1 = dlist_next (node1);
    }
    else
    {
      dlist_ins_next (output, dlist_tail (output), dlist_data (node2));
      count2--;
      node2 = dlist_next (node2);
    }
  }

  if (node1 == NULL)
    node1 = node2;

  for (; node1 != NULL; node1 = node1->next)
    dlist_ins_next (output, dlist_tail (output), dlist_data (node1));
  return 0;
};

int
print_clist (DList * list)
{
  DListElmt *element;
  DList *data;

  fprintf (stdout, "Number of significant Columns %d\n", dlist_size (list));

  dlist_iter (list, element)
  {
    data = (DList *) dlist_data (element);
    print_rle (data);
  }

  return 0;
}

void print_rle (DList * list)
{

  DListElmt *element;

  rlenode *data;
  int i;

  fprintf (stdout, "List size is %d\n", dlist_size (list));

  i = 0;

  dlist_iter (list, element)
  {

    data = (rlenode *) dlist_data (element);
    fprintf (stdout, "list[%03d]= %d %d %d\n", i, data->col, data->min,
             data->max);
    i++;
  }
}

int
find_rle (IntImage * image, int col, DList * list)
{
  int i, j;
  int run = 0;
  int min = 0;
  int max = 0;
  int count = 0;

  i = col * image->usize;
  min = i;
  max = i;
  rlenode *node;

  for (j = 0; j < image->usize; j++)
  {
    if (image->data[i + j])
    {
      max = j;
      if (run == 0)
      {
        min = j;
        run = 1;
      }
    }
    else
    {
      if (run)
      {
        node = (rlenode *) malloc (sizeof (rlenode));
        node->col = col;
        node->min = min;
        node->max = max;
        dlist_ins_prev (list, dlist_head (list), node);

        count++;
        run = 0;

      }
    }
  }
  if (run)
  {
    node = (rlenode *) malloc (sizeof (rlenode));
    node->col = col;
    node->min = min;
    node->max = max;

    dlist_ins_prev (list, dlist_head (list), node);

    count++;

  }
  return count;
}


int
rle_check_merge (DList * components, rlenode * run)
{


  DListElmt *cc;
  DListElmt *element;
  DList *cc_data;
  DList *newcomponent;
  DList *temp1;
  DList *temp2;

  rlenode *rle_data;

  DList *mergelist;

  mergelist = (DList *) malloc (sizeof (DList));
  dlist_init (mergelist, NULL);

  dlist_iter (components, cc)
  {
    cc_data = (DList *) dlist_data (cc);
    dlist_iter (cc_data, element)
    {
      rle_data = (rlenode *) dlist_data (element);

      if ((abs (rle_data->col - run->col) < 2) &&
          (intersect (run->min, run->max, rle_data->min, rle_data->max)))
      {

        dlist_ins_prev (mergelist, dlist_head (mergelist), cc);
        break;
      }
    }
  }


  if (dlist_size (mergelist))
  {

    temp1 = (DList *) malloc (sizeof (DList));
    dlist_init (temp1, NULL);
    dlist_ins_prev (temp1, NULL, run);

    dlist_iter (mergelist, cc)
    {

      dlist_remove (components, (DListElmt *) dlist_data (cc),
                    (void **) &cc_data);

      temp2 = (DList *) malloc (sizeof (DList));
      dlist_init (temp2, NULL);

      merge_rle (temp1, (DList *) cc_data, temp2);
      dlist_destroy (temp1);
      //experimental
      dlist_destroy ((DList *) cc_data);
      temp1 = temp2;

    }
    dlist_ins_prev (components, dlist_head (components), temp1);

  }
  else
  {
    newcomponent = (DList *) malloc (sizeof (DList));
    dlist_init (newcomponent, NULL);
    dlist_ins_prev (newcomponent, NULL, run);
    dlist_ins_prev (components, dlist_head (components), newcomponent);

  }

  return 0;

}

int
merge (DList * comp_list, DList * rle_list)
{
  DListElmt *node;
  dlist_iter (rle_list, node)
  {
    rle_check_merge (comp_list, (rlenode *) dlist_data (node));
  }
  return 0;
}

int
sublabel_image (DList * rlelist, int label, IntImage * image)
{
  int i;
  DListElmt *node;
  rlenode *rle_node;

  dlist_iter (rlelist, node)
  {
    rle_node = (rlenode *) dlist_data (node);
    for (i = rle_node->min; i < rle_node->max + 1; i++)
      image->data[index (i, rle_node->col, image->usize)] = label;
  };
  return 0;
}

int
label_image (DList * components, IntImage * image)
{
  int counter = 0;
  DListElmt *element;

  init_intimage (image);
  dlist_iter (components, element)
  {
    counter++;
    sublabel_image ((DList *) dlist_data (element), counter, image);
  }
  return 0;
}

int
free_clist (DList * components)
{
  DListElmt *element;
  DListElmt *node;
  DList *sublist;


  dlist_iter (components, element)
  {
    sublist = (DList *) dlist_data (element);
    dlist_iter (sublist, node)
    {
      free ((rlenode *) dlist_data (node));
    }

  }
  dlist_destroy (components);

  return 0;
}


int
bwlabel (IntImage * image, IntImage * output)
{
  int i;
  int count = 0;
  int size;
  IntImage *debug;

  DList components;             // this is the list of components
  DList *collist;
  DList *newlist;

  DListElmt *node;


  debug = new_image (image);
  // this is an explicit cast,
  // I should probably write this using a decent c++ version at some point of time
  dlist_init (&components, (void (*)(void *)) dlist_destroy);

  for (i = 0; i < image->vsize; i++)
  {
    collist = (DList *) malloc (sizeof (DList));
    dlist_init (collist, NULL);
    count += find_rle (image, i, collist);

    if (dlist_size (collist))
    {
      if (dlist_size (&components))
        merge (&components, collist);
      else
      {
        dlist_iter (collist, node)
        {
          newlist = (DList *) malloc (sizeof (DList));
          dlist_init (newlist, NULL);
          dlist_ins_prev (newlist, NULL, dlist_data (node));
          dlist_ins_prev (&components, dlist_head (&components), newlist);
        }
      }
    }
    else
      free (collist);
  }

  label_image (&components, output);
  size = dlist_size (&components);

  free_clist (&components);

  return size;
}
