#include "probe.h"
#include "unistd.h"
#include "ctype.h"

// filename functions

int
merge_subsets (IntImage * newmask,
               IntImage * oldmask,
               int nregions,
               int old_nregions, DList * light_array, DList * old_light_array)
{
  int i, j, k;
  bool subset;
  DListElmt *node;
  int *label = NULL;

  if (old_nregions)
  {
    label = (int *) malloc (sizeof (int) * (old_nregions + 1));

    for (i = 0; i < old_nregions + 1; i++)
      label[i] = 1;
  };

  printf ("Subsetting : ");
  for (i = 1; i <= nregions; i++)
  {
    printf ("*");
    dlist_init (light_array + i, NULL);
    k = 0;
    for (j = 1; j <= old_nregions; j++)
    {
      subset = 0;
      if (dlist_size (old_light_array + j) && (label[j]))
        subset = subset_intimage (oldmask, newmask, j, i);

      if (subset == 1)
      {
        if (label[j] == 0)
        {
          printf ("subset reuse :%d:%d\n", i, j);
        }

        label[j] = 0;

        dlist_iter (old_light_array + j, node)
        {
          k++;
          dlist_append (light_array + i, dlist_data (node));
          ((Light *) dlist_data (node))->index = k;
        }
      }

    }
  }
  printf ("\n");
  return 0;
}

int
process_hdr (char *probename,
             int lambda_n,
             double alpha,
             double beta,
             int nlights,
             int usize,
             int vsize, double dw, char *asciiname, char *binaryname)
{


  ProbeData *probe;
  ProbeData *patch;

  IntImage *mask;
  IntImage *cmask;
  IntImage *oldmask;
  IntImage *labelimage;

  DoubleArray *sarray;

  int i, j, k, l, n;

  int nregions = 0;
  int old_nregions = 0;
  float threshold;
  double svalue = 0.0;

  double is = 0.0;
  int isamples;
  double lambda;

  int light_count = 0;
  int total_light_count = 0;

  char dname[100];
  float t, r, s;

  DList *light_array = NULL;
  DList *old_light_array = NULL;


  probe = (ProbeData *) malloc (sizeof (ProbeData));
  probe->usize = usize;
  probe->vsize = vsize;
  probe->nsamples = usize * vsize;
  alloc_probe (probe);

  patch = (ProbeData *) malloc (sizeof (ProbeData));
  patch->usize = usize;
  patch->vsize = vsize;


  mask = (IntImage *) malloc (sizeof (IntImage));
  mask->usize = usize;
  mask->vsize = vsize;
  mask->size = usize * vsize;
  alloc_intimage (mask);

  cmask = new_image (mask);
  labelimage = new_image (mask);

  oldmask = new_image (mask);
  init_intimage (oldmask);


  // read the probe data
  if (read_probe (probename, probe) == false)
    abort ();


  printf ("Calculating S Value..\n");

  sarray = calculate_svalue (probe, -10000, alpha, beta, dw);
  svalue = 0.0;
  for (i = 0; i <= sarray->size; i++)
    svalue += sarray->data[i];


  free (sarray->data);
  free (sarray);

  printf ("S Value : %f \n", svalue);

  for (i = 0; i < lambda_n + 1; i++)
  {
    r = 0.15;
    s = 0.1;
    lambda = float (lambda_n - i);
    if (i == lambda_n)
      lambda = -1000;

    threshold = find_mask (probe, lambda, mask);
    clean_mask (mask, cmask);
    nregions = bwlabel (cmask, labelimage);
    printf ("Number of regions : %f:%d\n", lambda, nregions);

    light_array = (DList *) malloc (sizeof (DList) * (nregions + 1));

    merge_subsets (labelimage,
                   oldmask,
                   nregions, old_nregions, light_array, old_light_array);


    light_count = 0;

    for (j = 1; j <= nregions; j++)
    {

      isamples = 0;
      if (i == lambda_n)
      {
        // this can probably be simplified
        isamples = (int) (nlights - dlist_size (light_array + j));
      }
      else
      {
        is = patchs (probe, labelimage, alpha, beta, dw, j);

        if (nlights * is / svalue < 1.0)
        {
          is = patchs (probe, labelimage, alpha, 0.0, dw, j);
          if (nlights * is / svalue > 1.0)
            if (!dlist_size (light_array + j))
            {
              printf ("Adding a small bright sample\n");
              isamples = 1;
            }
        }
        else
        {
          isamples =
            (int) floor (nlights * is / svalue -
                         dlist_size (light_array + j));
        }
      }

      if (isamples < 0)
      {
        printf ("Negative samples : %d\n", isamples);
        isamples = 0;
      }

      if (isamples)
      {
        n = 0;
        for (k = 0; k < probe->nsamples; k++)
        {
          if (labelimage->data[k] == j)
            n++;
        }

        if (n < dlist_size (light_array + j) + isamples)
        {
          printf ("ERROR : patch %d:%d has %d elements\n", i,j, n);
          printf ("ERROR : %d : %d : %d\n", n, dlist_size (light_array + j),
                  isamples);
          abort ();
        };

        printf ("[%d:%d:%d] : %d=>[%d:%1.6f:%d]\n", i, j, nregions, n,
                dlist_size (light_array + j), is / svalue, isamples);
        patch->nsamples = n;
        alloc_probe (patch);
        l = 0;
        for (k = 0; k < probe->nsamples; k++)
        {
          if (labelimage->data[k] == j)
          {
            patch->map[l].r = probe->map[k].r;
            patch->map[l].g = probe->map[k].g;
            patch->map[l].b = probe->map[k].b;
            patch->map[l].u = probe->map[k].u;
            patch->map[l].v = probe->map[k].v;
            l++;
          }
        }
        shmoys (patch, isamples, light_array + j);

        free (patch->map);
      }
      light_count += isamples;
    }

    total_light_count += light_count;
    printf ("Light Count : %d:%d:%d\n", i, light_count, total_light_count);

    // destory old_light_array;
    free (old_light_array);
    old_light_array = light_array;
    old_nregions = nregions;
    copy_intimage (labelimage, oldmask);
  }

  if (asciiname != NULL)
    lights_ascii_writer (light_array, nregions, total_light_count, asciiname);

  if (binaryname != NULL)
    lights_binary_writer (light_array,
                          nregions, total_light_count, binaryname);

  free_intimage (labelimage);
  free_intimage (cmask);
  free_intimage (mask);
  free_probe (probe);
  return 0;
}

int
main (int argc, char **argv)
{
  int usize = 0;
  int vsize = 0;
  int lambda_n = 1;
  double alpha = 1.0;
  double beta = 0.25;
  int nlights = 0;
  double dw = 0.01;

  char *mapname = NULL;
  char *binaryname = NULL;
  char *asciiname = NULL;
  int c;
  // start parsing arguments

  opterr = 0;

  while ((c = getopt (argc, argv, "ho:b:d:p:q:n:s:w:")) != -1)
    switch (c)
    {
    case 'h':
      printf ("Structured Importance Sampling V3.5\n");
      printf
        ("stratify -d <number of steps> -s <size> -w <dw> -n <nlights> -o <asciifile> -b <binaryfile> <mapfilename>\n");
      return -1;
      break;
    case 'o':
      asciiname = optarg;
      break;
    case 'b':
      binaryname = optarg;
      break;
    case 'd':
      // lambda_n
      lambda_n = atoi (optarg);
      break;
    case 'n':
      nlights = atoi (optarg);
      break;
    case 'p':
      alpha = atof (optarg);
      break;
    case 'q':
      beta = atof (optarg);
      break;
    case 's':
      usize = atoi (optarg);
      vsize = usize;
      break;
    case 'w':
      dw = atof (optarg);
      break;
    case '?':
      if (isprint (optopt))
        fprintf (stderr, "Unknown option `-%c'.\n", optopt);
      else
        fprintf (stderr, "Unkown option character `\\x%x'.\n", optopt);
      return 1;
    default:
      abort ();
    }

  if (optind < argc)
    mapname = argv[optind];
  else
  {
    fprintf (stderr, "No mapname found\n");
    abort ();
  }
  // start error checks

  if (usize == 0)
  {
    printf ("Invalid or no value for mapsize specified\n");
    abort ();
  }

  if (lambda_n < 0)
  {
    printf ("lambda_n < 1 : invalid or no value specified\n");
    abort ();
  }


  printf ("Input Options \n");
  printf ("--------------------------------------------------------------\n");

  printf ("map name   : %s\n", mapname);
  printf ("steps      : %d\n", lambda_n + 1);
  printf ("nsamples   : %d\n", nlights);
  printf ("alpha/beta : %g/%g \n", alpha, beta);
  printf ("usize      : %d \n", usize);
  printf ("vsize      : %d \n", vsize);
  printf ("omega_0    : %f \n", dw);
  printf ("\n");
  printf ("Output Options \n");
  printf
    ("---------------------------------------------------------------\n");
  printf ("ascii lights filename  : %s\n", asciiname);
  printf ("binary lights filename : %s\n", binaryname);


  process_hdr (mapname,
               lambda_n,
               alpha,
               beta,
               nlights, 
               usize, 
               vsize, 
               dw, 
               asciiname, 
               binaryname);
  return 0;
}
