#include "probe.h"

void alloc_probe (ProbeData * probe)
{
  probe->map = (ProbeEntry *) malloc (sizeof (ProbeEntry) * probe->nsamples);
}

void free_probe (ProbeData * probe)
{
  free (probe->map);
  free (probe);
}

void get_pos (ProbeEntry * node, double *pos)
{
  float p, q;
  float theta, phi;
  p = node->u;
  q = node->v;
  theta = atan2 (q, p);
  phi = M_PI * sqrt (p * p + q * q);

  pos[0] = cos (theta) * sin (phi);
  pos[2] = sin (theta) * sin (phi);
  pos[1] = cos (phi);
}



float
get_area (ProbeEntry * entry)
{
  float r;
  float area;
  r = sqrt (entry->u * entry->u + entry->v * entry->v + EPS);
  if (r > 1.0)
    r = 1.0;

  area = 4 * M_PI * sin (M_PI * r) / r;
  if (area < 0.0)
    printf ("AREA error : %f %f : %f\n", entry->u, entry->v, area);
  return area;
};


double
patch_area (ProbeData * patch, int i)
{
  return (double) get_area (patch->map + i) / (patch->usize * patch->vsize);
}

bool read_probe (char *filename, ProbeData * probe)
{

  int i;
  int u, v;
  FILE *fp;
  float *is;

  printf ("Reading ProbeData : %s\n", filename);
  fp = fopen (filename, "r");

  if (!fp)
  {
    printf ("Cannot open \"%s\"\n", filename);
    return false;
  }
  printf ("Reading %d samples\n", probe->nsamples);

  is = (float *) malloc (sizeof (float) * 3 * probe->nsamples);
  fread (is, sizeof (float), 3 * probe->nsamples, fp);

  u = 0;
  v = probe->vsize - 1;
  printf ("Number of samples: %d\n", probe->nsamples);
  printf ("usize : %d\n", probe->usize);
  printf ("vsize : %d\n", probe->vsize);
  for (i = 0; i < probe->nsamples; i++)
  {

    probe->map[i].r = is[3 * i];
    probe->map[i].g = is[3 * i + 1];
    probe->map[i].b = is[3 * i + 2];
    probe->map[i].u = (2.0 * u) / probe->usize - 1;
    probe->map[i].v = (2.0 * v) / probe->vsize - 1;

    u++;
    if (u == probe->usize)
    {
      u = 0;
      v--;
      if (v < 0)
        v = probe->vsize - 1;
    }

  };
  return true;
}
