#include "probe.h"
#include "dlist.h"


inline double
norm (double *x)
{
  return sqrt (x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);
}


inline double
dists2 (double *x, double *y)
{
  double a;
  a = (x[0] * y[0] + x[1] * y[1] + x[2] * y[2]) /
    (sqrt ((x[0] * x[0] + x[1] * x[1] + x[2] * x[2]) *
           (y[0] * y[0] + y[1] * y[1] + y[2] * y[2])) + EPS);
  if (a > 1.0)
    return 0.0;
  else if (a < -1.0)
    return M_PI;
  return acos (a);

}

inline double
dists (double *x, double *y)
{
  // monotonic comparison
  return (1.1 -  (x[0] * y[0] + x[1] * y[1] + x[2] * y[2]) /
    (sqrt ((x[0] * x[0] + x[1] * x[1] + x[2] * x[2]) *
           (y[0] * y[0] + y[1] * y[1] + y[2] * y[2])) + EPS));
}


inline double
dot (double *x, double *y)
{
  return (x[1] * y[1] + x[2] * y[2] + x[0] * y[0]);
}

inline double
normalize (double *x)
{
  double n;
  n = norm (x);
  x[0] = x[0] / n;
  x[1] = x[1] / n;
  x[2] = x[2] / n;
  return 0;
}

void shmoys(ProbeData * patch, int number_of_lights, DList * lightlist)
{
  int i;

  double pos[3];
  double c;
  Light *newlight;
  int nlights;

  double max_dist = 0.0;
  int max_index;

  double newdist;
  double tarea;

  int pixelcount;

  DListElmt *old_index = NULL;

  int old_size = 0;

  DListElmt *element;
  Light *node;

  old_size = dlist_size (lightlist);

  int* label = (int *) malloc (sizeof (int) * patch->nsamples);
  double* dist = (double *) malloc (sizeof (double) * patch->nsamples);
  double* posarray = (double *) malloc (sizeof (double) * 3 * patch->nsamples);


  pos[0] = 0.0;
  pos[1] = 0.0;
  pos[2] = 0.0;

  nlights = 1;
  old_index = dlist_head (lightlist);

  for (i = 0; i < patch->nsamples; i++)
    get_pos (patch->map + i, posarray + 3 * i);

  printf ("Clustering : ");

  if (old_size)
  {
    printf ("+");

    newlight = (Light *) dlist_data (old_index);
    old_index = dlist_next (old_index);
    newlight->old = 0;
  }
  else
  {
    printf (".");
    max_dist = -1;
    max_index = 0;

    for (i = 0; i < patch->nsamples; i++)
    {
      c = (patch->map[i].r + patch->map[i].g + patch->map[i].b);
      if (c > max_dist)
      {
        max_dist = c;
        max_index = i;
      }
    };

    pos[0] = posarray[3 * max_index];
    pos[1] = posarray[3 * max_index + 1];
    pos[2] = posarray[3 * max_index + 2];
    normalize (pos);


    newlight = (Light *) malloc (sizeof (Light));
    newlight->pos[0] = pos[0];
    newlight->pos[1] = pos[1];
    newlight->pos[2] = pos[2];
    newlight->n = max_index;
    newlight->old = 0;
    dlist_append (lightlist, newlight);
  }

  newlight->index = nlights;

  for (i = 0; i < patch->nsamples; i++)
  {
    label[i] = nlights;
    dist[i] = dists (posarray + 3 * i, newlight->pos);
  }

  while (nlights < (number_of_lights + old_size))
  {
    nlights++;

    if (nlights <= old_size)
    {
      newlight = (Light *) dlist_data (old_index);
      old_index = dlist_next (old_index);
      newlight->old = 1;
      printf ("+");
    }
    else
    {
      printf (".");
      max_dist = -1.0;
      max_index = -1;

      for (i = 0; i < patch->nsamples; i++)
      {
        if (max_dist < dist[i])
        {
          max_dist = dist[i];
          max_index = i;
        }
      }
      newlight = (Light *) malloc (sizeof (Light));
      newlight->pos[0] = posarray[3 * max_index];
      newlight->pos[1] = posarray[3 * max_index + 1];
      newlight->pos[2] = posarray[3 * max_index + 2];
      newlight->n = max_index;
      newlight->old = 0;
      dlist_append (lightlist, newlight);
    }

    newlight->index = nlights;
    for (i = 0; i < patch->nsamples; i++)
    {
      newdist = dists (posarray + 3 * i, newlight->pos);
      if (newdist < dist[i])
      {
        dist[i] = newdist;
        label[i] = nlights;
      }
    }
  }



  dlist_iter (lightlist, element)
  {
    node = (Light *) dlist_data (element);
    node->color[0] = 0.0;
    node->color[1] = 0.0;
    node->color[2] = 0.0;
    node->area = 0.0;
    pixelcount = 0;
    for (i = 0; i < patch->nsamples; i++)
    {
      if (label[i] == node->index)
      {
        pixelcount++;
        tarea = patch_area (patch, i);
        node->area += (double) tarea;
        node->color[0] += (double) (tarea * patch->map[i].r);
        node->color[1] += (double) (tarea * patch->map[i].g);
        node->color[2] += (double) (tarea * patch->map[i].b);
      }
    }
    if (!pixelcount)
    {
      printf ("pixel count : %d:%d:%f \n", node->index, pixelcount,
              node->area);
      for (i = 0; i < patch->nsamples; i++)
      {
        printf ("ALERT : %d:%d:%d:%f:%f:%f:%f:%f:%f:%f\n", node->old, i,
                node->n, dists (posarray + 3 * i, node->pos), node->pos[0],
                node->pos[1], node->pos[1], posarray[3 * i],
                posarray[3 * i + 1], posarray[3 * i + 2]);

      }
    }
    node->color[0] = node->color[0] / node->area;
    node->color[1] = node->color[1] / node->area;
    node->color[2] = node->color[2] / node->area;

  }
  printf ("\n");
  free (label);
  free (dist);
  free (posarray);
}
