#include "probe.h"

DoubleArray *
calculate_svalue (ProbeData * probe,
                  double lambda, double alpha, double beta, double dw)
{
  double threshold = 0.0;
  int j;
  int nregions = 0;

  IntImage *mask;
  IntImage *cmask;

  DoubleArray *result;

  mask = (IntImage *) malloc (sizeof (IntImage));
  mask->usize = probe->usize;
  mask->vsize = probe->vsize;
  mask->size = probe->nsamples;
  alloc_intimage (mask);

  cmask = new_image (mask);

  threshold = find_mask (probe, lambda, mask);
  printf ("Thresholding : %f \n", threshold);
  clean_mask (mask, cmask);
  nregions = bwlabel (cmask, mask);

  printf ("number of regions :%d\n", nregions);
  result = (DoubleArray *) malloc (sizeof (DoubleArray));
  result->size = nregions;
  result->data = (double *) malloc (sizeof (double) * (nregions + 1));

  for (j = 0; j <= nregions; j++)
  {
    result->data[j] = patchs (probe, mask, alpha, beta, dw, j);
  }

  free_intimage (cmask);
  free_intimage (mask);
  return result;
}


double
patchs (ProbeData * probe,
        IntImage * mask, double alpha, double beta, double dw, int label)
{
  int i;
  double tmp;
  double area = 0.0;
  double L = 0.0;
  int count = 0;
  double value;

  for (i = 0; i < probe->nsamples; i++)
  {

    if (mask->data[i] == label)
    {
      tmp = patch_area (probe, i);
      area += tmp;
      L += (probe->map[i].r + probe->map[i].g + probe->map[i].b) * tmp / 3.0;
      count++;
    }
  }

  if (area > dw)
    area = dw;

  value = pow (L, alpha) * pow (area, beta);

  return value;
}
